"""Autoupdate older conda dependencies in the requirements section."""

import collections
import itertools
import re
import xml.etree.ElementTree as ET
from typing import (
    Any,
    DefaultDict,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    TYPE_CHECKING,
    Union,
)
from xml.etree.ElementTree import ElementTree

import packaging.version
import requests
import yaml
from bioblend import toolshed
from bioblend.toolshed import ToolShedInstance
from galaxy.tool_util.deps import conda_util

import planemo.conda
from planemo.io import (
    error,
    info,
)

if TYPE_CHECKING:
    from planemo.cli import PlanemoCliContext
    from planemo.galaxy.config import LocalGalaxyConfig
    from planemo.runnable import Runnable

AUTOUPDATE_TOOLSHED_URL = "https://toolshed.g2.bx.psu.edu"


def find_macros(xml_tree: ElementTree) -> List[Any]:
    """
    Get macros from the XML tree
    """
    macros = []
    for macro_import in xml_tree.iter("import"):
        macros.append(macro_import.text)
    return macros


def get_requirements(xml_tree: ElementTree) -> Tuple[Dict[str, Dict[str, Optional[str]]], Optional[str]]:
    """
    Get requirements from the XML tree
    """
    requirements = {}
    main_req = None
    for requirement in xml_tree.iter("requirement"):
        if requirement.attrib.get("version") == "@TOOL_VERSION@":
            main_req = requirement.text
        else:
            assert requirement.text
            requirements[requirement.text] = {
                "tag": ET.tostring(requirement, encoding="unicode").strip(),
                "text": requirement.attrib.get("version"),
            }
    return requirements, main_req


def get_tokens(xml_tree: ElementTree) -> Dict[str, Dict[str, Optional[str]]]:
    """
    Get tokens from the XML tree
    """
    tokens = {}
    for token in xml_tree.iter("token"):
        tokens[token.attrib["name"]] = {"tag": ET.tostring(token, encoding="unicode").strip(), "text": token.text}
    return tokens


def check_conda(package_name: str, ctx: "PlanemoCliContext", **kwds) -> str:
    """
    Get the most up-to-date conda version for a package.
    """
    conda_context = planemo.conda.build_conda_context(ctx, **kwds)
    if not conda_context.is_conda_installed():
        # check directly via Anaconda API
        r = requests.get("https://api.anaconda.org/search", params={"name": package_name})
        search_results = itertools.chain.from_iterable(
            n["versions"] for n in r.json() if n["name"] == package_name and n["owner"] in kwds["conda_ensure_channels"]
        )
        return sorted(search_results, key=packaging.version.parse, reverse=True)[0]

    target = conda_util.CondaTarget(package_name)
    best_search_results = conda_util.best_search_result(target, conda_context=conda_context)
    if best_search_results[0] is None:
        raise Exception(f"No conda package found for {package_name}")
    return best_search_results[0]["version"]


def update_xml(
    tool_path: str,
    xml_tree: ElementTree,
    tags_to_update: List[Dict[str, str]],
    wrapper_version_token: Optional[Union[int, str]],
    is_macro: bool = False,
) -> None:
    """
    Write modified XML to tool_path
    """

    def update_token(xml_text, tag, token_value):
        new_tag = f">{token_value}<".join(re.split(">.*<", tag))
        return re.sub(tag, new_tag, xml_text)

    def update_requirement(xml_text, tag, requirement_value):
        new_tag = f'version="{requirement_value}"'.join(re.split('version=".*"', tag))
        return re.sub(tag, new_tag, xml_text)

    with open(tool_path, "r+", newline="") as f:
        xml_text = f.read()
        for tag_to_update in tags_to_update:
            if tag_to_update["type"] == "token":
                xml_text = update_token(xml_text, tag_to_update["tag"], tag_to_update["value"])
            if tag_to_update["type"] == "requirement":
                xml_text = update_requirement(xml_text, tag_to_update["tag"], tag_to_update["value"])
        if wrapper_version_token == 0 and not is_macro:
            # i.e. @VERSION_SUFFIX@ not specified so update the version directly in the tool tag
            tool_tag = re.sub(
                'version="@TOOL_VERSION@.*?"',
                'version="@TOOL_VERSION@+galaxy0"',
                re.findall('<tool .*version="@TOOL_VERSION@.*">', xml_text)[0],
            )
            xml_text = re.sub('<tool .*version="@TOOL_VERSION@.*">', tool_tag, xml_text)
        f.seek(0)
        f.truncate()
        f.write(xml_text)


def create_requirement_dict(
    xml_files: Dict[str, ElementTree], skip_reqs: List[str]
) -> Tuple[Dict[str, Dict[str, Dict[str, Optional[str]]]], Optional[Tuple[str, str]]]:
    """
    Create dict with requirements and find main requirement
    """
    requirements = {}
    main_req = None
    for k, v in xml_files.items():
        file_reqs, file_main_req = get_requirements(v)
        requirements[k] = {k: v for k, v in file_reqs.items() if k not in skip_reqs}
        if file_main_req:
            if main_req:
                error("Multiple requirements use the token @TOOL_VERSION@!")
            main_req = (file_main_req, k)
    if not main_req:
        error("No requirement uses the token @TOOL_VERSION@!")
    return requirements, main_req


def create_token_dict(
    ctx: "PlanemoCliContext", xml_files: Dict[str, ElementTree], main_req: Tuple[str, str], **kwds
) -> Tuple[
    Dict[str, Dict[str, Dict[str, Optional[str]]]], DefaultDict[str, List[Dict[str, str]]], Optional[str], Optional[str]
]:
    """
    Create dict with relevant tokens and check conda requirements for main
    """
    tokens: Dict[str, Dict[str, Dict[str, Optional[str]]]] = {}
    current_main_req, updated_main_req = None, None
    xml_to_update = collections.defaultdict(list)
    for k, v in xml_files.items():
        tokens[k] = get_tokens(v)
        # check if it is @TOOL_VERSION@ and if so do check_conda
        if "@TOOL_VERSION@" in tokens[k]:
            current_main_req = tokens[k]["@TOOL_VERSION@"]["text"]
            updated_main_req = check_conda(main_req[0], ctx, **kwds)
            if current_main_req:
                tag = tokens[k]["@TOOL_VERSION@"]["tag"]
                assert tag is not None
                xml_to_update[k].append({"type": "token", "tag": tag, "value": updated_main_req})

    return tokens, xml_to_update, current_main_req, updated_main_req


def perform_required_update(
    ctx: "PlanemoCliContext",
    xml_files: Dict[str, ElementTree],
    tool_path: str,
    requirements: Dict[str, Dict[str, Dict[str, Optional[str]]]],
    tokens: Dict[str, Dict[str, Dict[str, Optional[str]]]],
    xml_to_update: DefaultDict[str, List[Dict[str, str]]],
    wrapper_version_token: Optional[Union[int, str]],
    **kwds,
) -> Set[str]:
    """
    Carry out the update, if requirements are out-of-date
    """
    # check all requirements
    for k, v in requirements.items():
        for req in v:
            req_check = check_conda(req, ctx, **kwds)
            # print(req_check, v[req]['text'])
            if req_check != v[req]["text"]:
                tag = v[req]["tag"]
                assert tag is not None
                xml_to_update[k].append({"type": "requirement", "tag": tag, "value": req_check})

    # check all tokens, if wrapper_version_token exists
    if wrapper_version_token:
        for k, v in tokens.items():
            if isinstance(wrapper_version_token, str) and wrapper_version_token in v:
                tag = v[wrapper_version_token]["tag"]
                assert tag is not None
                xml_to_update[k].append({"type": "token", "tag": tag, "value": "0"})

    # finally, update each file separately
    for k, et in xml_files.items():
        update_xml(k, et, xml_to_update[k], wrapper_version_token, is_macro=(k != tool_path))
    info(f"Tool {tool_path} successfully updated.")
    return set(xml_files)


def autoupdate_tool(ctx: "PlanemoCliContext", tool_path: str, modified_files: Set[Any], **kwds) -> Optional[Set[str]]:
    """
    Autoupdate an XML file
    """
    modified_files = modified_files or set()
    # create a dict of all files that need editing - wrapper plus macros
    xml_files = {tool_path: ET.parse(tool_path)}

    # get name of token which defines the wrapper version; if just an integer, None
    versions = xml_files[tool_path].getroot().attrib.get("version")
    if versions:
        versions = versions.split("+galaxy")
        if versions[0] != "@TOOL_VERSION@":
            error("Tool version does not contain @TOOL_VERSION@ as required by autoupdate.")
            return None
        elif len(versions) == 1:
            wrapper_version_token = None
        else:
            if versions[1][0] == versions[1][-1] == "@":
                wrapper_version_token = versions[1]
            else:
                wrapper_version_token = 0  # assume an int, reset to 0
    else:
        wrapper_version_token = None

    # add macros to xml_files
    for macro in find_macros(xml_files[tool_path]):
        macro_path = "/".join(tool_path.split("/")[:-1] + [macro])
        xml_files[macro_path] = ET.parse(macro_path)

    requirements, main_req = create_requirement_dict(xml_files, kwds.get("skip_requirements", "").split(","))
    if main_req is None:
        return None
    tokens, xml_to_update, current_main_req, updated_main_req = create_token_dict(ctx, xml_files, main_req, **kwds)

    if current_main_req == updated_main_req and not (modified_files & set(xml_files)):
        info(f"No updates required or made to {tool_path}.")
        return None  # end here if no update needed

    if kwds.get("dry_run"):
        error(
            f"Update required to {tool_path}! Tool main requirement has version {current_main_req}, newest conda version is {updated_main_req}"
        )
        return None

    else:
        info(f"Updating {tool_path.split('/')[-1]} from version {current_main_req} to {updated_main_req}")
        return perform_required_update(
            ctx, xml_files, tool_path, requirements, tokens, xml_to_update, wrapper_version_token, **kwds
        )


def _update_wf(config: "LocalGalaxyConfig", workflow_id: str, instance: bool = False) -> None:
    """
    Recursively update a workflow, including subworkflows
    """
    wf = config.user_gi.make_get_request(
        f"{config.user_gi.url}/workflows/{workflow_id}", params={"instance": instance}
    ).json()
    for step in wf.get("steps", {}).values():
        if step["type"] == "subworkflow":
            # update subworkflows before the main workflow
            _update_wf(config, step["workflow_id"], instance=True)
    config.user_gi.workflows.refactor_workflow(wf["id"], actions=[{"action_type": "upgrade_all_steps"}])


def outdated_tools(
    ctx: "PlanemoCliContext", wf_dict: Dict[str, Any], ts: ToolShedInstance
) -> Dict[str, Dict[str, str]]:
    def check_tool_step(step, ts):  # return a dict with current and newest tool version, in case they don't match
        if not step["tool_id"].startswith(AUTOUPDATE_TOOLSHED_URL[8:]):
            return {}  # assume a built in tool
        try:
            repos = ts.repositories._get(params={"tool_ids": step["tool_id"]})
        except Exception:
            ctx.log(f"The ToolShed returned an error when searching for the most recent version of {step['tool_id']}")
            return {}
        base_id = "/".join(step["tool_id"].split("/")[:-1])
        tool_ids_found = {
            tool["guid"] for repo in repos.values() if type(repo) == dict for tool in repo.get("tools", [])
        }
        updated_tool_id = sorted(
            {tool_id for tool_id in tool_ids_found if f"{base_id}/" in tool_id},
            key=lambda n: packaging.version.parse(n),
        )[-1]
        if step["tool_id"] != updated_tool_id:
            return {base_id: {"current": step["tool_id"], "updated": updated_tool_id}}
        else:
            return {}

    outdated_tool_dict = {}
    steps = wf_dict["steps"].values() if type(wf_dict["steps"]) == dict else wf_dict["steps"]
    for step in steps:
        if step.get("type", "tool") == "tool" and not step.get("run", {}).get("class") == "GalaxyWorkflow":
            outdated_tool_dict.update(check_tool_step(step, ts))
        elif step.get("type") == "subworkflow":  # GA SWF
            outdated_tool_dict.update(outdated_tools(ctx, step["subworkflow"], ts))
        elif step.get("run", {}).get("class") == "GalaxyWorkflow":  # gxformat2 SWF
            outdated_tool_dict.update(outdated_tools(ctx, step["run"], ts))
        else:
            continue
    return outdated_tool_dict


def get_tools_to_update(
    ctx: "PlanemoCliContext", workflow: "Runnable", tools_to_skip: List[Any]
) -> Dict[str, Dict[str, str]]:
    # before we run the autoupdate, we check the tools against the toolshed to see if there
    # are any new versions. This saves spinning up Galaxy and installing the tools if there
    # is nothing to do, and also allows us to collect a list of the tools which need updating
    with open(workflow.path) as f:
        wf_dict = yaml.load(f, Loader=yaml.SafeLoader)

    ts = toolshed.ToolShedInstance(url=AUTOUPDATE_TOOLSHED_URL)
    tools_to_update = outdated_tools(ctx, wf_dict, ts)
    return {tool: versions for tool, versions in tools_to_update.items() if tool not in tools_to_skip}


def autoupdate_wf(ctx: "PlanemoCliContext", config: "LocalGalaxyConfig", wf: "Runnable") -> Dict[str, Any]:
    workflow_id = config.workflow_id_for_runnable(wf)
    _update_wf(config, workflow_id)
    return config.user_gi.workflows.export_workflow_dict(workflow_id)


def fix_workflow_ga(original_wf: Dict[str, Any], updated_wf: Dict[str, Any]) -> Dict[str, Any]:
    # the Galaxy refactor action can't do everything right now... some manual fixes here
    # * bump release number if present
    # * order steps numerically, leave everything else sorted as in the original workflow
    # * recurse over subworkflows
    edited_wf = original_wf.copy()
    updated_wf_steps = collections.OrderedDict(sorted(updated_wf["steps"].items(), key=lambda item: int(item[0])))
    edited_wf["steps"] = updated_wf_steps
    # check release; bump if it exists
    if edited_wf.get("release"):
        release = [int(n) for n in edited_wf["release"].split(".")]
        release[-1] += 1
        edited_wf["release"] = ".".join([str(n) for n in release])
    # iterate over the steps
    for step in edited_wf["steps"]:
        # recurse over subworkflows
        if edited_wf["steps"][step].get("type") == "subworkflow":
            edited_wf["steps"][step]["subworkflow"] = fix_workflow_ga(
                edited_wf["steps"][step]["subworkflow"], updated_wf["steps"][step]["subworkflow"]
            )
    return edited_wf


def fix_workflow_gxformat2(original_wf: Dict[str, Any], updated_wf: Dict[str, Any]) -> Dict[str, Any]:
    # does the same as fix_workflow_ga for gxformat2
    edited_wf = original_wf.copy()
    # check release; bump if it exists
    if edited_wf.get("release"):
        release = [int(n) for n in edited_wf["release"].split(".")]
        release[-1] += 1
        edited_wf["release"] = ".".join([str(n) for n in release])
    # iterate over the steps
    for step_index, step in enumerate(edited_wf["steps"]):
        # recurse over subworkflows
        if step.get("run", {}).get("class") == "GalaxyWorkflow":  # subworkflow
            step["run"] = fix_workflow_gxformat2(
                step["run"], updated_wf["steps"][str(step_index + len(original_wf["inputs"]))]["subworkflow"]
            )
        # fix tool_id and content_id to march tool_version
        elif updated_wf["steps"][str(step_index + len(original_wf["inputs"]))]["type"] == "tool":
            if (
                updated_wf["steps"][str(step_index + len(original_wf["inputs"]))]
                .get("tool_id", "")
                .startswith(AUTOUPDATE_TOOLSHED_URL[8:])
            ):
                step["tool_version"] = updated_wf["steps"][str(step_index + len(original_wf["inputs"]))]["tool_version"]
                step["tool_id"] = updated_wf["steps"][str(step_index + len(original_wf["inputs"]))]["tool_id"]
                step["content_id"] = updated_wf["steps"][str(step_index + len(original_wf["inputs"]))]["content_id"]

    return edited_wf



#!/usr/bin/env python

import ast
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.version_info < (3, 7):
    sys.stderr.write("ERROR: planemo requires at least Python version 3.7\n")
    sys.exit(1)

# Allow installer to turn off dependency on lxml by setting the environment variable
# PLANEMO_REQUIRE_LXML to "0". lxml should be considered optional if xmllint is
# available on the PATH - but python doesn't really provide me a fantastic way to
# express that.
DEFAULT_PLANEMO_REQUIRE_LXML = 1
PLANEMO_REQUIRE_LXML = os.environ.get("PLANEMO_REQUIRE_LXML", "%d" % DEFAULT_PLANEMO_REQUIRE_LXML) != "0"

SOURCE_DIR = "planemo"

_version_re = re.compile(r"__version__\s+=\s+(.*)")


with open("%s/__init__.py" % SOURCE_DIR, "rb") as f:
    init_contents = f.read().decode("utf-8")

    def get_var(var_name):
        pattern = re.compile(r"%s\s+=\s+(.*)" % var_name)
        match = pattern.search(init_contents).group(1)
        return str(ast.literal_eval(match))

    version = get_var("__version__")
    PROJECT_NAME = get_var("PROJECT_NAME")
    PROJECT_URL = get_var("PROJECT_URL")
    PROJECT_AUTHOR = get_var("PROJECT_AUTHOR")
    PROJECT_EMAIL = get_var("PROJECT_EMAIL")

TEST_DIR = "tests"
PROJECT_DESCRIPTION = (
    "Command-line utilities to assist in building tools for the Galaxy project (http://galaxyproject.org/)."
)
PACKAGES = [
    "planemo",
    "planemo.cwl",
    "planemo.commands",
    "planemo.database",
    "planemo.engine",
    "planemo.galaxy",
    "planemo.galaxy.test",
    "planemo.linters",
    "planemo.reports",
    "planemo.shed",
    "planemo.shed2tap",
    "planemo.test",
    "planemo.training",
    "planemo.xml",
]
ENTRY_POINTS = """
    [console_scripts]
    planemo=planemo.cli:planemo
"""
PACKAGE_DATA = {
    "planemo": [
        "xml/xsd/repository_dependencies.xsd",
        "xml/xsd/tool_dependencies.xsd",
        "reports/*",
        "scripts/*",
    ],
}
PACKAGE_DIR = {
    SOURCE_DIR: SOURCE_DIR,
}

with open("README.rst") as fh:
    readme = fh.read()
with open("HISTORY.rst") as fh:
    history = fh.read().replace(".. :changelog:", "")

if os.path.exists("requirements.txt"):
    with open("requirements.txt") as fh:
        requirements = [r for r in fh.read().split("\n") if ";" not in r]
    if not PLANEMO_REQUIRE_LXML:
        requirements.remove("lxml")
else:
    # In tox, it will cover them anyway.
    requirements = []

test_requirements = [
    # TODO: put package test requirements here
]


setup(
    name=PROJECT_NAME,
    version=version,
    description=PROJECT_DESCRIPTION,
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    author=PROJECT_AUTHOR,
    author_email=PROJECT_EMAIL,
    url=PROJECT_URL,
    packages=PACKAGES,
    entry_points=ENTRY_POINTS,
    package_data=PACKAGE_DATA,
    package_dir=PACKAGE_DIR,
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords="planemo",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Topic :: Software Development",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Testing",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite=TEST_DIR,
    tests_require=test_requirements,
)



import os
import subprocess
import sys
import tarfile
import zipfile
from ftplib import all_errors as FTPErrors  # tuple of exceptions
from typing import List
from urllib.error import URLError
from urllib.request import urlretrieve
from xml.etree import ElementTree

from galaxy.util import unicodify

TOOLSHED_MAP = {
    "toolshed": "https://toolshed.g2.bx.psu.edu",
    "testtoolshed": "https://testtoolshed.g2.bx.psu.edu",
}


class Dependencies:
    """Base class for parsing Tool Shed dependency files."""

    def __init__(
        self,
        dependencies_file,
        repo=None,
        package_factory=None,
    ):
        if package_factory is None:
            package_factory = BasePackage
        self.repo = repo
        self.root = ElementTree.parse(dependencies_file).getroot()
        packages = []
        dependencies = []
        package_els = self.root.findall("package")
        assert package_els is not None
        for package_el in package_els:
            install_els = package_el.findall("install")
            readme_els = package_el.findall("readme")
            if len(readme_els) > 0:
                readme = readme_els[0].text
            else:
                readme = None
            assert len(install_els) in (0, 1)
            if len(install_els) == 1:
                install_el = install_els[0]
                package = package_factory(self, package_el, install_el, readme=readme)
                packages.append(package)
            else:
                repository_el = package_el.find("repository")
                if repository_el is None:
                    message = f"no repository in package el for {repo}"
                    raise AssertionError(message)
                dependency = Dependency(self, package_el, repository_el)
                dependencies.append(dependency)

        self.packages = packages
        self.dependencies = dependencies

    def single_package(self):
        return len(self.packages) == 1

    def __repr__(self):
        return f"Dependencies[for_repo={self.repo}]"


class Repo:
    def __init__(self, **kwds):
        for key, value in kwds.items():
            setattr(self, key, value)

    def recipe_base_name(self):
        owner = self.owner.replace("-", "")
        name = self.name
        name = name.replace("_", "").replace("-", "")
        base = f"{owner}_{name}"
        return base

    @staticmethod
    def from_xml(elem):
        tool_shed_url = elem.attrib.get("toolshed", None)
        if tool_shed_url and ("testtoolshed" in tool_shed_url):
            prefix = "testtoolshed"
        else:
            prefix = "toolshed"
        prior = elem.attrib.get("prior_installation_required", False)
        return Repo(
            prefix=prefix,
            name=elem.attrib["name"],
            owner=elem.attrib["owner"],
            tool_shed_url=tool_shed_url,
            changeset_revision=elem.attrib.get("changeset_revision", None),
            prior_installation_required=prior,
        )

    @staticmethod
    def from_api(prefix, repo_json):
        return Repo(
            prefix=prefix,
            name=repo_json["name"],
            owner=repo_json["owner"],
            tool_shed_url=TOOLSHED_MAP[prefix],
        )

    def get_file(self, path):
        try:
            url = f"{self.tool_shed_url}/repos/{self.owner}/{self.name}/raw-file/tip/{path}"
            path, headers = urlretrieve(url)
            return path
        except Exception as e:
            print(e)
            return None

    def __repr__(self):
        return f"Repository[name={self.name},owner={self.owner}]"


class Dependency:
    def __init__(self, dependencies, package_el, repository_el):
        self.dependencies = dependencies
        self.package_el = package_el
        self.repository_el = repository_el
        self.repo = Repo.from_xml(repository_el)

    def __repr__(self):
        return (
            f"Dependency[package_name={self.package_el.attrib['name']},version={self.package_el.attrib['version']},"
            f"dependent_package={self.repository_el.attrib['name']}]"
        )


class BasePackage:
    def __init__(self, dependencies, package_el, install_el, readme):
        self.dependencies = dependencies
        self.package_el = package_el
        self.install_el = install_el
        self.readme = readme
        self.all_actions = self.get_all_actions()
        self.no_arch_option = self.has_no_achitecture_install()

    def get_all_actions(self):
        action_or_group = self.install_el[0]
        parsed_actions = []
        if action_or_group.tag == "actions":
            parsed_actions.append(self.parse_actions(action_or_group))
        elif action_or_group.tag == "actions_group":
            actions_els = action_or_group.findall("actions")
            assert actions_els is not None
            for actions in actions_els:
                parsed_actions.append(self.parse_actions(actions))
            action_els = action_or_group.findall("action")
            assert action_els is not None
            for action in action_els:
                for parsed_a in parsed_actions:
                    parsed_a.actions.append(self.parse_action(action))
        return parsed_actions

    def has_no_achitecture_install(self):
        all_actions = self.all_actions
        if len(all_actions) < 2:
            return False
        else:
            last_action = all_actions[-1]
            return (not last_action.architecture) and (not last_action.os)

    def has_explicit_set_environments(self):
        all_actions = self.all_actions
        for actions in all_actions:
            for action in actions.actions:
                if action.explicit_variables:
                    return True
        return False

    def has_multiple_set_environments(self):
        all_actions = self.all_actions
        for actions in all_actions:
            count = 0
            for action in actions.actions:
                if action.explicit_variables:
                    count += 1
            if count > 1:
                return True
        return False

    def parse_actions(self, actions):
        os = actions.attrib.get("os", None)
        architecture = actions.get("architecture", None)
        action_els = actions.findall("action")
        assert action_els is not None
        parsed_actions = list(map(self.parse_action, action_els))
        action_packages = []
        for package in actions.findall("package"):
            action_packages.append(self.parse_action_package(package))
        return Actions(parsed_actions, os, architecture, action_packages)

    def parse_action_package(self, elem):
        name = elem.attrib["name"]
        version = elem.attrib["version"]
        repo = Repo.from_xml(elem.find("repository"))
        return ActionPackage(name, version, repo)

    def parse_action(self, action):
        return BaseAction.from_elem(action, package=self)

    def __repr__(self):
        actions = self.all_actions
        return (
            f"Install[name={self.package_el.attrib['name']},version={self.package_el.attrib['version']},"
            f"dependencies={self.dependencies},actions={actions}]"
        )


class Actions:
    def __init__(self, actions, os=None, architecture=None, action_packages=[]):
        self.os = os
        self.architecture = architecture
        self.actions = actions or []
        self.action_packages = action_packages

    def first_download(self):
        for action in self.actions:
            if action.action_type in ["download_by_url", "download_file"]:
                return action
        return None

    def downloads(self):
        actions = []
        for action in self.actions:
            if action.action_type in ["download_by_url", "download_file"]:
                actions.append(action)
        return actions

    def __repr__(self):
        platform = ""
        if self.os or self.architecture:
            platform = f"os={self.os},arch={self.architecture},"
        return f"Actions[{platform}{map(str, self.actions)}]"

    def _indent_extend(self, target, new_entries, indent="    "):
        for line in new_entries:
            target.append(indent + line)

    def to_bash(self):
        # Use self.os.title() to match "Linux" or "Darwin" in bash where case matters:
        if self.os and self.architecture:
            condition = f'("{self.os.title()}" == `uname`) && ("{self.architecture}" == `arch`)'
        elif self.os:
            condition = f'"{self.os.title()}" == `uname`'
        elif self.architecture:
            condition = f'"{self.architecture}" == `arch`'
        else:
            condition = None

        install_cmds = []
        env_cmds = []

        if condition:
            # Conditional actions block
            install_cmds = [
                "#" + "-" * 60,
                f"if [[ $specifc_action_done == 0 && {condition} ]]",
                "then",
                f'    echo "Platform-specific action for os={self.os}, arch={self.architecture}"',
            ]
            env_cmds = install_cmds[:]
            # TODO - Refactor block indentation?
            for action in self.actions:
                i_cmds, e_cmds = action.to_bash()
                self._indent_extend(install_cmds, i_cmds)
                self._indent_extend(env_cmds, e_cmds)
            # If we run the action, do not want to run any later actions!
            install_cmds.extend(["    specifc_action_done=1", "fi"])
            env_cmds.extend(["    specifc_action_done=1", "fi"])
        else:
            # Non-specific default action...
            install_cmds = [
                "#" + "-" * 60,
                "if [[ $specifc_action_done == 0 ]]",
                "then",
                '    echo "Non-platform-specific actions"',
            ]
            env_cmds = install_cmds[:]
            for action in self.actions:
                i_cmds, e_cmds = action.to_bash()
                self._indent_extend(install_cmds, i_cmds)
                self._indent_extend(env_cmds, e_cmds)
            install_cmds.append("fi")
            env_cmds.append("fi")
        return install_cmds, env_cmds


class ActionPackage:
    def __init__(self, name, version, repo):
        self.name = name
        self.version = version
        self.repo = repo


class BaseAction:
    _keys = []  # type: List[str]
    action_type = None  # type: str

    def __repr__(self):
        return f"Action[type={self.action_type}]"

    def same_as(self, other):
        if self._keys != other._keys:
            return False
        else:
            for key in self._keys:
                if getattr(self, key) != getattr(other, key):
                    return False

            return True

    def parse_action_repo(self, elem):
        repo_elem = elem.find("repository")
        repo = Repo.from_xml(repo_elem)
        self.repo = repo

    def parse_package_elems(self, elem):
        package_els = elem.findall("package")
        packages = []
        assert package_els is not None
        for package_el in package_els:
            packages.append(package_el.text)
        self.packages = packages

    @classmethod
    def from_elem(cls, elem, package):
        type = elem.attrib["type"]
        action_class = actions_by_type[type]
        return action_class(elem)

    def to_bash(self):
        """Return lists of bash shell commands to execute this action.

        This method is be implemented by each sub-class, and will
        return two list of strings (for ``dep_install.sh`` and
        ``env.sh`` respectively).
        """
        raise NotImplementedError(f"No to_bash defined for {self!r}")


def _tar_folders(filename):
    with tarfile.open(filename, "r", errorlevel=0) as archive:
        folders = set()
        for i in archive.getmembers():
            if i.isdir():
                folders.add(i.name.rstrip("/"))
            else:
                folders.add(os.path.split(i.name)[0])
        return list(folders)


def _zip_folders(filename):
    archive = zipfile.ZipFile(filename, "r")
    return list({i.filename.rstrip("/") for i in archive.infolist() if i.filename.endswith("/")})


def _common_prefix(folders):
    common_prefix = ""
    if len(folders) == 1:
        common_prefix = list(folders)[0]
    else:
        common_prefix = os.path.commonprefix(folders)
        assert not os.path.isabs(common_prefix), folders
    return common_prefix


def _cache_download(url, filename, sha256sum=None):
    """Returns local path to cached copy of URL using given filename."""
    cache = os.environ.get("DOWNLOAD_CACHE", "./download_cache/")
    # TODO - expose this as a command line option

    if not os.path.isdir(cache):
        os.mkdir(cache)

    local = os.path.join(cache, filename)

    if not os.path.isfile(local):
        # Must download it...
        try:
            # TODO - log this nicely...
            sys.stderr.write(f"Downloading {url} to {local!r}\n")
            urlretrieve(url, local)
        except URLError:
            # Most likely server is down, could be bad URL in XML action:
            raise RuntimeError(f"Unable to download {url}")
        except FTPErrors:
            # Most likely server is down, could be bad URL in XML action:
            raise RuntimeError(f"Unable to download {url}")

        # Verifying the checksum is slow, only do this on a fresh
        # download. Assume locally cached files are already OK.
        if sha256sum:
            # TODO - log this nicely...
            sys.stderr.write(f"Verifying checksum for {filename}\n")
            filehash = subprocess.check_output(["shasum", "-a", "256", local])[0:64].strip()
            filehash = unicodify(filehash)
            if filehash != sha256sum:
                raise RuntimeError(f"Checksum failure for {local}, got {filehash!r} but wanted {sha256sum!r}")

    return local


def _determine_compressed_file_folder(url, downloaded_filename, target_filename=None, sha256sum=None):
    """Determine how to decompress the file & its directory structure.

    Returns a list of shell commands. Consider this example where the
    folder to change to cannot be guessed from the tar-ball filename:

        $ curl -o "ncbi-blast-2.2.30+-ia32-linux.tar.gz" \
        "ftp://ftp.ncbi.nlm.nih.gov/blast/executables/blast+/2.2.30/ncbi-blast-2.2.30+-ia32-linux.tar.gz"
        $ tar -zxvf ncbi-blast-2.2.30+-ia32-linux.tar.gz
        $ cd ncbi-blast-2.2.30+

    Here it would return:

        ['tar -zxvf ncbi-blast-2.2.30+-ia32-linux.tar.gz', 'cd ncbi-blast-2.2.30+']

    If not cached, this function will download the file to the
    $DOWNLOAD_CACHE folder, and then open it / decompress it in
    order to find common folder prefix used.  This will also verify
    how to decompress the file, and the checksum if given.
    """
    answer = []

    local = _cache_download(url, downloaded_filename, sha256sum)

    if not target_filename:
        target_filename = downloaded_filename

    if tarfile.is_tarfile(local):
        folders = _tar_folders(local)
        if target_filename.endswith((".tar.gz", ".tgz")):
            answer.append(f"tar -zxvf {target_filename}")
        elif target_filename.endswith(".tar.bz2"):
            answer.append(f"tar -jxvf {target_filename}")
        elif target_filename.endswith(".tar"):
            answer.extend(f"tar -xvf {target_filename}")
        else:
            # Quite possibly this file doesn't need decompressing,
            # but until we've tested lots of real world tool_dependencies.xml
            # files I'd like to check these cases to confirm this.
            raise NotImplementedError(f"How to decompress tar file {target_filename}?")
    elif zipfile.is_zipfile(local):
        if target_filename.endswith(".jar"):
            # Do not decompress!
            return answer
        folders = _zip_folders(local)
        answer.append(f"unzip {target_filename}")
    elif target_filename.endswith(".dmg"):
        # Do not decompress!
        return answer
    else:
        # No compression? Leave as it is?
        raise NotImplementedError(f"What kind of compression is {local} using?")

    common_prefix = _common_prefix(folders)
    if common_prefix:
        answer.append(f'cd "{common_prefix}"')

    return answer


def _commands_and_downloaded_file(url, target_filename=None, sha256sum=None):
    # We preserve the filename from the URL in the cache.
    # i.e. We do NOT use the target_filename in the cache.
    # This because some Galaxy recipes normalise platform specific downloads
    # to use a single target filename, which would therefore break checksums etc
    # e.g. tests/data/repos/package_1/tool_dependencies.xml
    downloaded_filename = os.path.split(url)[-1]
    if "?" in downloaded_filename:
        downloaded_filename = downloaded_filename[: downloaded_filename.index("?")]
    if "#" in downloaded_filename:
        downloaded_filename = downloaded_filename[: downloaded_filename.index("#")]

    if not target_filename:
        target_filename = downloaded_filename

    # Curl is present on Mac OS X, can we assume it will be on Linux?
    # Cannot assume that wget will be on Mac OS X.
    answer = [
        f'if [[ -f "{target_filename}" ]]',
        "then",
        f'    echo "Reusing existing {target_filename}"',
        f'elif [[ -f "$DOWNLOAD_CACHE/{downloaded_filename}" ]]',
        "then",
        f'    echo "Reusing cached {downloaded_filename}"',
        f'    cp "$DOWNLOAD_CACHE/{downloaded_filename}" "{target_filename}"',
        "else",
        f'    echo "Downloading {downloaded_filename}"',
        f'    curl -L -o "$DOWNLOAD_CACHE/{downloaded_filename}" "{url}"',
        f'    cp "$DOWNLOAD_CACHE/{downloaded_filename}" "{target_filename}"',
    ]
    if sha256sum:
        # This is inserted into the if-else for a fresh download only.
        # Note double space between checksum and filename:
        answer.append(f'    echo "{sha256sum}  {target_filename}" | shasum -a 256 -c -')
    answer.append("fi")

    return answer, downloaded_filename


def _commands_to_download_and_extract(url, target_filename=None, sha256sum=None):
    answer, downloaded_filename = _commands_and_downloaded_file(url, target_filename, sha256sum)
    # Now should we unpack the tar-ball etc?
    answer.extend(_determine_compressed_file_folder(url, downloaded_filename, target_filename, sha256sum))
    return answer, []


class DownloadByUrlAction(BaseAction):
    action_type = "download_by_url"
    _keys = ["url"]

    def __init__(self, elem):
        self.url = elem.text.strip()
        assert self.url
        self.sha256sum = elem.attrib.get("sha256sum", None)
        self.target_filename = elem.attrib.get("target_filename", None)

    def to_bash(self):
        # See class DownloadByUrl in Galaxy,
        # lib/tool_shed/galaxy_install/tool_dependencies/recipe/step_handler.py
        return _commands_to_download_and_extract(self.url, self.target_filename, self.sha256sum)


class DownloadFileAction(BaseAction):
    action_type = "download_file"
    _keys = ["url", "extract"]

    def __init__(self, elem):
        self.url = elem.text.strip()
        self.extract = asbool(elem.attrib.get("extract", False))
        self.sha256sum = elem.attrib.get("sha256sum", None)
        self.target_filename = elem.attrib.get("target_filename", None)

    def to_bash(self):
        if self.extract:
            return _commands_to_download_and_extract(self.url, self.target_filename, self.sha256sum)
        else:
            commands, downloaded_file = _commands_and_downloaded_file(self.url, self.target_filename, self.sha256sum)
            return commands, []


class DownloadBinary(BaseAction):
    action_type = "download_binary"
    _keys = ["url_template", "target_directory"]

    def __init__(self, elem):
        self.url_template = elem.text
        assert self.url_template
        self.target_directory = elem.get("target_directory", None)


class ShellCommandAction(BaseAction):
    action_type = "shell_command"
    _keys = ["command"]

    def __init__(self, elem):
        self.command = elem.text

    def to_bash(self):
        # Galaxy would run each action from the same temp
        # working directory - possible that tool_dependencies.xml
        # shell_command could change $PWD so reset this:
        return ["pushd . > /dev/null", self.command, "popd > /dev/null"], []


class TemplateShellCommandAction(BaseAction):
    action_type = "template_command"
    _keys = ["language", "command"]

    def __init__(self, elem):
        self.command = elem.text
        self.language = elem.get("language", "cheetah").lower()
        assert self.command
        assert self.language == "cheetah"


class MoveFileAction(BaseAction):
    action_type = "move_file"
    _keys = ["move_file"]

    def __init__(self, elem):
        self.source = elem.find("source").text
        self.destination = elem.find("destination").text

    def to_bash(self):
        return [f"mv {self.source} {self.destination}"], []


class MoveDirectoryFilesAction(BaseAction):
    action_type = "move_directory_files"
    _keys = ["source_directory", "destination_directory"]

    def __init__(self, elem):
        source_directory = elem.find("source_directory").text
        destination_directory = elem.find("destination_directory").text
        self.source_directory = source_directory
        self.destination_directory = destination_directory

    def to_bash(self):
        return [f"mv {self.source_directory}/* {self.destination_directory}/"], []


class SetEnvironmentAction(BaseAction):
    action_type = "set_environment"
    _keys = ["variables"]

    def __init__(self, elem):
        variables = []
        var_els = elem.findall("environment_variable")
        assert var_els is not None
        for ev_elem in var_els:
            var = SetVariable(ev_elem)
            variables.append(var)
        self.variables = variables
        assert self.variables

    def to_bash(self):
        answer = []
        for var in self.variables:
            # Expand $INSTALL_DIR here?
            if var.action == "set_to":
                answer.append(f"export {var.name}={var.raw_value}")
            elif var.action == "prepend_to":
                answer.append(f"export {var.name}={var.raw_value}:${var.name}")
            elif var.action == "append_to":
                answer.append(f"export {var.name}=${var.name}:{var.raw_value}")
            else:
                raise ValueError(f"Undefined environment variable action {var.action!r}")
        return answer, answer  # Actions needed in env.sh here!


class ChmodAction(BaseAction):
    action_type = "chmod"
    _keys = ["mods"]

    def __init__(self, elem):
        mods = []
        file_els = elem.findall("file")
        assert file_els is not None
        for mod_elem in file_els:
            mod = {}
            mod["mode"] = mod_elem.attrib["mode"]
            mod["target"] = mod_elem.text
            mods.append(mod)
        self.mods = mods
        assert self.mods

    def to_bash(self):
        return [f"chmod {m['mode']} {m['target']}" for m in self.mods], []


class MakeInstallAction(BaseAction):
    action_type = "make_install"
    _keys = []  # type: List[str]

    def __init__(self, elem):
        pass

    def to_bash(self):
        return ["make install"], []


class AutoconfAction(BaseAction):
    action_type = "autoconf"
    _keys = ["options"]

    def __init__(self, elem):
        self.options = elem.text

    def to_bash(self):
        if self.options:
            raise NotImplementedError("Options with action autoconf not implemented yet.")
        return ["./configure", "make", "make install"], []


class ChangeDirectoryAction(BaseAction):
    action_type = "change_directory"
    _keys = ["directory"]

    def __init__(self, elem):
        self.directory = elem.text
        assert self.directory

    def to_bash(self):
        return [f"cd {self.directory}"], []


class MakeDirectoryAction(BaseAction):
    action_type = "make_directory"
    _keys = ["directory"]

    def __init__(self, elem):
        self.directory = elem.text

    def to_bash(self):
        return [f"mkdir -p {self.directory}"], []


class SetupPerlEnvironmentAction(BaseAction):
    action_type = "setup_perl_environment"
    _keys = ["repo", "packages"]

    def __init__(self, elem):
        self.parse_action_repo(elem)
        self.parse_package_elems(elem)


class SetupRubyEnvironmentAction(BaseAction):
    action_type = "setup_ruby_environment"
    _keys = ["repo", "packages"]

    def __init__(self, elem):
        self.parse_action_repo(elem)
        self.parse_package_elems(elem)


class SetupPythonEnvironmentAction(BaseAction):
    action_type = "setup_python_environment"
    _keys = ["repo", "packages"]

    def __init__(self, elem):
        self.parse_action_repo(elem)
        self.parse_package_elems(elem)


class SetupVirtualenvAction(BaseAction):
    action_type = "setup_virtualenv"
    _keys = ["use_requirements_file", "python", "requirements"]

    def __init__(self, elem):
        use_reqs = elem.attrib.get("use_requirements_file", "True")
        self.use_requirements_file = asbool(use_reqs)
        self.python = elem.get("python", "python")
        self.requirements = elem.text or "requirements.txt"


class SetupREnvironmentAction(BaseAction):
    action_type = "setup_r_environment"
    _keys = ["repo", "packages"]

    def __init__(self, elem):
        self.parse_action_repo(elem)
        self.parse_package_elems(elem)


class SetEnvironmentForInstallAction(BaseAction):
    action_type = "set_environment_for_install"

    def __init__(self, elem):
        pass

    def to_bash(self):
        # TODO - How could we resolve/check the dependencies?
        return ['echo "WARNING: Assuming packages already installed!"'], []


class SetVariable:
    def __init__(self, elem):
        self.action = elem.attrib["action"]
        self.name = elem.attrib["name"]
        self.raw_value = elem.text


truthy = frozenset(["true", "yes", "on", "y", "t", "1"])
falsy = frozenset(["false", "no", "off", "n", "f", "0"])


def asbool(obj):
    if isinstance(obj, str):
        obj = obj.strip().lower()
        if obj in truthy:
            return True
        elif obj in falsy:
            return False
        else:
            raise ValueError(f"String is not true/false: {obj!r}")
    return bool(obj)


action_classes = [
    DownloadByUrlAction,
    DownloadFileAction,
    DownloadBinary,
    ShellCommandAction,
    TemplateShellCommandAction,
    MoveFileAction,
    MoveDirectoryFilesAction,
    SetEnvironmentAction,
    ChmodAction,
    MakeInstallAction,
    AutoconfAction,
    ChangeDirectoryAction,
    MakeDirectoryAction,
    SetupPerlEnvironmentAction,
    SetupRubyEnvironmentAction,
    SetupPythonEnvironmentAction,
    SetupVirtualenvAction,
    SetupREnvironmentAction,
    SetEnvironmentForInstallAction,
]

actions_by_type = dict(map(lambda c: (c.action_type, c), action_classes))



"""Module describing the planemo ``merge_test_reports`` command."""
import os

import click

from planemo import (
    io,
    options,
)
from planemo.cli import command_function
from planemo.galaxy.test.actions import merge_reports


@click.command("merge_test_reports")
@options.merge_test_json()
@options.tool_test_json("output_path")
@command_function
def cli(ctx, input_paths, output_path, **kwds):
    """Merge tool_test_output.json files from multiple runs."""
    for input_path in input_paths:
        if not os.path.exists(input_path):
            io.error("Failed to tool test json file at %s" % input_path)
            return 1
    merge_reports(input_paths=input_paths, output_path=output_path)



"""Module describing the planemo ``test_reports`` command."""
import datetime
import pathlib

import click

from planemo import (
    io,
    options,
)
from planemo.cli import command_function
from planemo.galaxy.test import (
    handle_reports,
    StructuredData,
)


@click.command("test_reports")
@options.tool_test_json()
@options.test_report_options()
@command_function
def cli(ctx, path, **kwds):
    """Generate human readable tool test reports.

    Creates reports in various formats  (HTML, text, markdown)
    from the structured test output (tool_test_output.json).
    """
    fname = pathlib.Path(path)
    if not fname.exists():
        io.error("Failed to tool test json file at %s" % path)
        return 1

    test_data = StructuredData(path)
    test_data.calculate_summary_data_if_needed()
    file_modication_datatime = datetime.datetime.fromtimestamp(fname.stat().st_mtime)
    kwds["file_modication_datatime"] = file_modication_datatime
    handle_reports(ctx, test_data.structured_data, kwds)



"""Module describing the planemo ``workflow_upload`` command."""
from collections import defaultdict
from pathlib import Path

import click
from gxformat2.yaml import ordered_load_path

from planemo import options
from planemo.cli import command_function
from planemo.github_util import create_release
from planemo.workflow_lint import find_workflow_descriptions


@click.command("workflow_upload")
@options.github_namespace()
@options.github_branch()
@options.dry_run()
@options.optional_tools_or_packages_arg(multiple=True)
@command_function
def cli(ctx, paths, namespace, dry_run, github_branch, **kwds):
    """Upload workflows to github organization."""
    owner = namespace
    for path in paths:
        path = Path(path).absolute()
        if path.is_dir():
            repo = path.name
        else:
            repo = path.parent.name
            path = path.parent
        versions = defaultdict(list)
        for workflow_file in find_workflow_descriptions(path):
            workflow = ordered_load_path(workflow_file)
            version = workflow.get("release")
            if not version:
                raise Exception(f"Must set a release version in workflow file '{workflow_file}'")
            versions[version].append(workflow_file)
            if len(versions) > 1:
                msg = ""
                for version, paths in versions.items():
                    msg = "{}version: {}\npaths: {}".format(msg, version, "\n".join(paths))
                raise Exception(f"All workflows in repository must have same version.\n{msg}")
        if versions:
            create_release(
                ctx,
                from_dir=path,
                target_dir=repo,
                owner=owner,
                repo=repo,
                version=version,
                dry_run=dry_run,
                branch=github_branch,
            )



"""Module describing the planemo ``shed_serve`` command."""
import click

from planemo import (
    io,
    options,
    shed,
)
from planemo.cli import command_function
from planemo.engine import engine_context
from planemo.galaxy.serve import sleep_for_serve
from planemo.runnable_resolve import install_args_list_to_runnables


@click.command("shed_serve")
@options.shed_read_options()
@options.galaxy_serve_options()
@click.option(
    "--skip_dependencies", is_flag=True, help="Do not install shed dependencies as part of repository installation."
)
@command_function
def cli(ctx, paths, **kwds):
    """Launch Galaxy with Tool Shed dependencies.

    This command will start a Galaxy instance configured to target the
    specified shed, find published artifacts (tools and dependencies)
    corresponding to command-line arguments and ``.shed.yml`` file(s),
    install these artifacts, and serve a Galaxy instances that can be
    logged into and explored interactively.
    """
    kwds["galaxy_skip_client_build"] = kwds.pop("skip_client_build", False)
    install_args_list = kwds["install_args_list"] = shed.install_arg_lists(ctx, paths, **kwds)
    runnables = install_args_list_to_runnables(ctx, install_args_list, kwds)
    with engine_context(ctx, **kwds) as engine:
        with engine.ensure_runnables_served(runnables) as config:
            io.info(f"Galaxy running with tools installed at {config.galaxy_url}")
            sleep_for_serve()



"""Module describing the planemo ``serve`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.galaxy import galaxy_serve
from planemo.runnable import for_paths
from planemo.tools import uris_to_paths


@click.command("serve")
@options.optional_tools_arg(multiple=True, allow_uris=True)
@options.galaxy_serve_options()
@options.enable_cwl_option()
@options.galaxy_cwl_root_option()
@command_function
def cli(ctx, uris, **kwds):
    """Launch Galaxy instance with specified tools.

    The Galaxy tool panel will include just the referenced tool or tools (by
    default all the tools in the current working directory) and the upload
    tool.

    planemo will search parent directories to see if any is a Galaxy instance
    - but one can pick the Galaxy instance to use with the ``--galaxy_root``
    option or force planemo to download a disposable instance with the
    ``--install_galaxy`` flag.

    ``planemo`` will run the Galaxy instance in an existing virtualenv if one
    exists in a ``.venv`` directory in the specified ``--galaxy_root``.
    Otherwise, the Galaxy instance will run in a clean virtualenv created in
    ``/tmp``.

    ``planemo`` uses temporarily generated config files and environment
    variables to attempt to shield this execution of Galaxy from manually
    launched runs against that same Galaxy root - but this may not be bullet
    proof yet, so please be careful and do not try this against a production
    Galaxy instance.
    """
    paths = uris_to_paths(ctx, uris)
    runnables = for_paths(paths)
    kwds["galaxy_skip_client_build"] = kwds.pop("skip_client_build", False)
    with galaxy_serve(ctx, runnables, **kwds):
        pass



"""Module describing the planemo ``workflow_test_init`` command."""
import os

import click
import yaml

from planemo import options
from planemo.cli import command_function
from planemo.galaxy.workflows import (
    get_workflow_from_invocation_id,
    job_template,
    new_workflow_associated_path,
    output_stubs_for_workflow,
)
from planemo.io import can_write_to_path


@click.command("workflow_test_init")
@options.required_workflow_arg()
@options.force_option()
@options.workflow_output_artifact()
@options.split_job_and_test()
@options.galaxy_url_option()
@options.galaxy_user_key_option()
@options.from_invocation()
@options.profile_option()
@command_function
def cli(ctx, workflow_identifier, output=None, split_test=False, **kwds):
    """Initialize a Galaxy workflow test description for supplied workflow.

    Be sure to lint your workflow with ``workflow_lint`` before calling this
    to ensure inputs and outputs comply with best practices that make workflow
    testing easier.
    """
    if kwds["from_invocation"]:
        if not os.path.isdir("test-data"):
            ctx.log("Creating test-data directory.")
            os.makedirs("test-data")
        path_basename = get_workflow_from_invocation_id(
            workflow_identifier, kwds["galaxy_url"], kwds["galaxy_user_key"]
        )
    else:
        path_basename = os.path.basename(workflow_identifier)
    job = job_template(workflow_identifier, **kwds)
    if output is None:
        output = new_workflow_associated_path(path_basename if kwds["from_invocation"] else workflow_identifier)
    job_output = new_workflow_associated_path(
        path_basename if kwds["from_invocation"] else workflow_identifier, suffix="job1"
    )
    if not can_write_to_path(output, **kwds):
        ctx.exit(1)

    test_description = [
        {
            "doc": "Test outline for %s" % path_basename,
            "job": job,
            "outputs": output_stubs_for_workflow(workflow_identifier, **kwds),
        }
    ]
    if split_test:
        job_output = new_workflow_associated_path(
            path_basename if kwds["from_invocation"] else workflow_identifier, suffix="job1"
        )
        if not can_write_to_path(job_output, **kwds):
            ctx.exit(1)

        test_description[0]["job"] = os.path.basename(job_output)
        with open(job_output, "w") as f_job:
            yaml.dump(job, f_job, sort_keys=False)
    with open(output, "w") as f:
        yaml.dump(test_description, f, sort_keys=False)



"""Utilities to help linting various targets."""

import os
from urllib.request import urlopen

import requests
from galaxy.tool_util.lint import LintContext

from planemo.io import error
from planemo.shed import find_urls_for_xml
from planemo.xml import validation


def build_lint_args(ctx, **kwds):
    """Handle common report, error, and skip linting arguments."""
    report_level = kwds.get("report_level", "all")
    fail_level = kwds.get("fail_level", "warn")
    skip = kwds.get("skip", None)
    if skip is None:
        skip = ctx.global_config.get("lint_skip", "")
        if isinstance(skip, list):
            skip = ",".join(skip)

    skip_types = [s.strip() for s in skip.split(",")]
    lint_args = dict(
        level=report_level,
        fail_level=fail_level,
        skip_types=skip_types,
    )
    return lint_args


def setup_lint(ctx, **kwds):
    """Prepare lint_args and lint_ctx to begin linting a target."""
    lint_args = kwds.get("lint_args", None) or build_lint_args(ctx, **kwds)
    lint_ctx = LintContext(lint_args["level"])
    return lint_args, lint_ctx


def handle_lint_complete(lint_ctx, lint_args, failed=False):
    """Complete linting of a target and decide exit code."""
    if not failed:
        failed = lint_ctx.failed(lint_args["fail_level"])
    if failed:
        error("Failed linting")
    return 1 if failed else 0


def lint_dois(tool_xml, lint_ctx):
    """Find referenced DOIs and check they have valid with https://doi.org."""
    dois = find_dois_for_xml(tool_xml)
    for publication in dois:
        is_doi(publication, lint_ctx)


def find_dois_for_xml(tool_xml):
    dois = []
    for element in tool_xml.getroot().findall("citations"):
        for citation in list(element):
            if citation.tag == "citation" and citation.attrib.get("type", "") == "doi":
                dois.append(citation.text)
    return dois


def is_doi(publication_id, lint_ctx):
    """Check if dx.doi knows about the ``publication_id``."""
    base_url = "https://doi.org"
    if publication_id is None:
        lint_ctx.error("Empty DOI citation")
        return
    publication_id = publication_id.strip()
    doiless_publication_id = publication_id.split("doi:", 1)[-1]
    if not doiless_publication_id:
        lint_ctx.error("Empty DOI citation")
        return
    url = f"{base_url}/{doiless_publication_id}"
    r = requests.get(url)
    if r.status_code == 200:
        if publication_id != doiless_publication_id:
            lint_ctx.error("%s is valid, but Galaxy expects DOI without 'doi:' prefix" % publication_id)
        else:
            lint_ctx.info("%s is a valid DOI" % publication_id)
    elif r.status_code == 404:
        lint_ctx.error("%s is not a valid DOI" % publication_id)
    else:
        lint_ctx.warn("dx.doi returned unexpected status code %d" % r.status_code)


def lint_xsd(lint_ctx, schema_path, path):
    """Lint XML at specified path with supplied schema."""
    name = lint_ctx.object_name or os.path.basename(path)
    validator = validation.get_validator(require=True)
    validation_result = validator.validate(schema_path, path)
    if not validation_result.passed:
        msg = "Invalid XML found in file: %s. Errors [%s]"
        msg = msg % (name, validation_result.output)
        lint_ctx.error(msg)
    else:
        lint_ctx.info("File validates against XML schema.")


def lint_urls(root, lint_ctx):
    """Find referenced URLs and verify they are valid."""
    urls, docs = find_urls_for_xml(root)

    # This is from Google Chome on macOS, current at time of writing:
    BROWSER_USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36"

    def validate_url(url, lint_ctx, user_agent=None):
        is_valid = True
        if url.startswith("http://") or url.startswith("https://"):
            if user_agent:
                headers = {"User-Agent": user_agent, "Accept": "*/*"}
            else:
                headers = None
            r = None
            try:
                r = requests.get(url, headers=headers, stream=True)
                r.raise_for_status()
                next(r.iter_content(1000))
            except Exception as e:
                if r is not None and r.status_code == 429:
                    # too many requests
                    pass
                if r is not None and r.status_code in [403, 503] and "cloudflare" in r.text:
                    # CloudFlare protection block
                    pass
                else:
                    is_valid = False
                    lint_ctx.error(f"Error '{e}' accessing {url}")
        else:
            try:
                with urlopen(url) as handle:
                    handle.read(100)
            except Exception as e:
                is_valid = False
                lint_ctx.error(f"Error '{e}' accessing {url}")
        if is_valid:
            lint_ctx.info("URL OK %s" % url)

    for url in urls:
        validate_url(url, lint_ctx)
    for url in docs:
        validate_url(url, lint_ctx, BROWSER_USER_AGENT)


__all__ = (
    "build_lint_args",
    "handle_lint_complete",
    "lint_dois",
    "lint_urls",
    "lint_xsd",
)



"""Module contains code for gtdk: Galaxy training development kit."""

from planemo.io import info
from .topic import Topic
from .tutorial import Tutorial


class Training:
    """Class to describe a training."""

    def __init__(self, kwds):
        """Init an instance of Training."""
        self.kwds = kwds
        self.topics_dir = "topics"
        self.topic = Topic(parent_dir=self.topics_dir, name=kwds["topic_name"])
        self.galaxy_url = kwds["galaxy_url"] if "galaxy_url" in kwds else ""
        self.galaxy_api_key = kwds["galaxy_api_key"] if "galaxy_api_key" in kwds else ""
        self.tuto = None

    def init_training(self, ctx):
        """Create/update a topic/tutorial."""
        if not self.topic.exists():
            info("The topic %s does not exist. It will be created" % self.topic.name)
            self.topic.init_from_kwds(self.kwds)
            self.topic.create_topic_structure()

        if not self.kwds["tutorial_name"]:
            if self.kwds["slides"]:
                raise Exception("A tutorial name is needed to create the skeleton of a tutorial slide deck")
            if self.kwds["workflow"] or self.kwds["workflow_id"]:
                raise Exception("A tutorial name is needed to create the skeleton of the tutorial from a workflow")
            if self.kwds["zenodo_link"]:
                raise Exception("A tutorial name is needed to add Zenodo information")
        else:
            self.tuto = Tutorial(training=self, topic=self.topic)
            self.tuto.init_from_kwds(self.kwds)
            if not self.tuto.exists():
                info(f"The tutorial {self.tuto.name} in topic {self.topic.name} does not exist. It will be created.")
                self.tuto.create_tutorial(ctx)
        info("WARNING: Change the contributors/maintainers listed in the metadata of the new training ")
        info("before serving the website to fit the one listed in the CONTRIBUTORS.yaml file")

    def check_topic_init_tuto(self):
        """Check that the topic and tutorial are already there and retrieve them."""
        # check topic
        if not self.topic.exists():
            raise Exception("The topic %s does not exists. It should be created" % self.topic.name)
        self.topic.init_from_metadata()
        # initiate the tutorial
        self.tuto = Tutorial(training=self, topic=self.topic)
        self.tuto.init_from_existing_tutorial(self.kwds["tutorial_name"])
        if "datatypes" in self.kwds:
            self.tuto.datatype_fp = self.kwds["datatypes"]
        if "workflow" in self.kwds:
            self.tuto.init_wf_fp = self.kwds["workflow"]
        if "workflow_id" in self.kwds:
            self.tuto.init_wf_id = self.kwds["workflow_id"]

    def fill_data_library(self, ctx):
        """Fill a data library for a tutorial."""
        self.check_topic_init_tuto()
        # get the zenodo link
        z_link = ""
        update_metadata = False
        if self.tuto.zenodo_link != "":
            if self.kwds["zenodo_link"]:
                info("The data library and the metadata will be updated with the new Zenodo link")
                z_link = self.kwds["zenodo_link"]
                self.tuto.zenodo_link = z_link
                update_metadata = True
            else:
                info("The data library will be extracted using the Zenodo link in the metadata of the tutorial")
                z_link = self.tuto.zenodo_link
        elif self.kwds["zenodo_link"]:
            info("The data library will be created and the metadata will be filled with the new Zenodo link")
            z_link = self.kwds["zenodo_link"]
            self.tuto.zenodo_link = z_link
            update_metadata = True

        if z_link == "" or z_link is None:
            raise Exception(
                "A Zenodo link should be provided either in the metadata file or as argument of the command"
            )

        # extract the data library from Zenodo
        self.tuto.prepare_data_library_from_zenodo()

        # update the metadata
        if update_metadata:
            self.tuto.write_hands_on_tutorial(add_z_file_links=False)

    def generate_tuto_from_wf(self, ctx):
        """Generate the skeleton of a tutorial from a workflow."""
        self.check_topic_init_tuto()
        if self.tuto.has_workflow():
            info("Create tutorial skeleton from workflow")
            self.tuto.create_hands_on_tutorial(ctx)
            self.tuto.export_workflow_file()
        else:
            raise Exception(
                "A path to a local workflow or the id of a workflow on a running Galaxy instance should be provided"
            )



"""Module describing the planemo ``profile_list`` command."""

import click

from planemo.cli import command_function
from planemo.galaxy import profiles
from planemo.io import info


@click.command("profile_list")
@command_function
def cli(ctx, **kwds):
    """List configured profile names."""
    info("Looking for profiles...")
    profile_names = profiles.list_profiles(ctx, **kwds)
    for profile in profile_names:
        print(profile)
    info(f"{len(profile_names)} configured profiles are available.")



"""Module describing the planemo ``shed_download`` command."""
import sys

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function

target_path = click.Path(
    file_okay=True,
    writable=True,
    resolve_path=True,
)


@click.command("shed_download")
@options.shed_read_options()
@click.option(
    "--destination",
    default="shed_download.tar.gz",
    type=target_path,
    help="Destination pattern of tarball(s) to download - if this doesn't "
    "end in 'gz' it will be treated as a directory to extract tool "
    "contents into (defaults to shed_download.tar.gz). If multiple "
    "repositories are discovered in a .shed.yml file these will be "
    "created as shed_download_<name>.tar.gz by default for instance, "
    "simpler repositories will just be downloaded to the specified file.",
)
@command_function
def cli(ctx, paths, **kwds):
    """Download tool from Tool Shed into directory.

    Download a tool repository as a tarball from the tool shed and extract
    to the specified directory.
    """
    shed_context = shed.get_shed_context(ctx, read_only=True, **kwds)

    def download(realized_repository):
        return shed.download_tarball(ctx, shed_context, realized_repository, **kwds)

    exit_code = shed.for_each_repository(ctx, download, paths, **kwds)
    sys.exit(exit_code)



"""Module describing the planemo ``conda_env`` command."""

import click
from galaxy.tool_util.deps import conda_util

from planemo import options
from planemo.cli import command_function
from planemo.conda import (
    build_conda_context,
    collect_conda_targets,
)
from planemo.io import (
    error,
    ps1_for_path,
)

SOURCE_COMMAND = """
PRE_CONDA_PS1=$PS1
source %s %s
if [[ -n $BASH_VERSION ]]; then
    hash -r
elif [[ -n $ZSH_VERSION ]]; then
    rehash
else
echo 'Only bash and zsh are supported'
    return 1
fi
PS1="%s"
echo 'Deactivate environment with conda_env_deactivate'
alias conda_env_deactivate="source %s; %s"
"""


@click.command("conda_env")
@options.optional_tools_arg()
@options.conda_target_options()
# @options.skip_install_option()  # TODO
@command_function
def cli(ctx, path, **kwds):
    """Activate a conda environment for tool.

    Source the output of this command to activate a conda environment for this
    tool.

    \b
        $ . <(planemo conda_env seqtk_seq.xml)
        Deactivate environment with conda_env_deactivate
        (seqtk_seq_v6) $ which seqtk
        /home/planemo/miniconda2/envs/jobdepsDkzcjjfecc6d406196737781ff4456ec60975c137e04884e4f4b05dc68192f7cec4656/bin/seqtk
        (seqtk_seq_v6) $ conda_env_deactivate
        $

    """
    conda_context = build_conda_context(ctx, use_planemo_shell_exec=False, **kwds)
    conda_targets = collect_conda_targets(ctx, [path])
    installed_conda_targets = conda_util.filter_installed_targets(conda_targets, conda_context=conda_context)
    env_name, exit_code = conda_util.build_isolated_environment(
        installed_conda_targets, conda_context=conda_context, quiet=True
    )
    if exit_code:
        error("Failed to build environment for request.")
        return 1

    ps1 = ps1_for_path(path, base="PRE_CONDA_PS1")
    remove_env = f"{conda_context.conda_exec} env remove -y --name '{env_name}'"
    deactivate = conda_context.deactivate
    activate = conda_context.activate
    command = SOURCE_COMMAND % (activate, env_name, ps1, deactivate, remove_env)
    print(command)



"""Module describing the planemo ``workflow_lint`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.lint import build_lint_args
from planemo.workflow_lint import lint_workflow_artifacts_on_paths


@click.command("workflow_lint")
@options.optional_tools_or_packages_arg(multiple=True)
@options.report_level_option()
@options.report_xunit()
@options.fail_level_option()
@options.skip_option()
@command_function
def cli(ctx, paths, **kwds):
    """Check workflows for syntax errors and best practices."""
    # Unlike tools, lets just make this recursive by default.
    lint_args = build_lint_args(ctx, **kwds)
    exit_code = lint_workflow_artifacts_on_paths(
        ctx,
        paths,
        lint_args,
    )
    ctx.exit(exit_code)



"""Utilities for interacting with git using planemo abstractions."""

import os
import subprocess
from typing import (
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
)

from galaxy.util import unicodify

from planemo import io

if TYPE_CHECKING:
    from planemo.cli import PlanemoCliContext


def git_env_for(path: str) -> Dict[str, str]:
    """Setup env dictionary to target specified git repo with git commands."""
    env = os.environ.copy()
    env.update({"GIT_WORK_DIR": path, "GIT_DIR": os.path.join(path, ".git")})
    return env


def ls_remote(ctx: "PlanemoCliContext", remote_repo: str) -> Dict[str, str]:
    """Return a dictionary with refs as key and commits as value."""
    commits_and_refs = io.communicate(
        ["git", "ls-remote", remote_repo],
        stdout=subprocess.PIPE,
    )[0]
    return dict(line.split()[::-1] for line in commits_and_refs.decode("utf-8").splitlines())


def init(ctx: "PlanemoCliContext", repo_path: str) -> None:
    env = git_env_for(repo_path)
    io.communicate(["git", "init"], env=env)


def add(ctx: "PlanemoCliContext", repo_path: str, file_path: str) -> None:
    env = git_env_for(repo_path)
    io.communicate(["git", "add", os.path.relpath(file_path, repo_path)], env=env, cwd=repo_path)


def commit(ctx: "PlanemoCliContext", repo_path: str, message: str = "") -> None:
    env = git_env_for(repo_path)
    io.communicate(["git", "commit", "-m", message], env=env)


def push(
    ctx: "PlanemoCliContext",
    repo_path: str,
    to: Optional[str] = None,
    branch: Optional[str] = None,
    force: bool = False,
) -> None:
    env = git_env_for(repo_path)
    cmd = ["git", "push"]
    if force:
        cmd += ["--force"]
    if to and branch:
        cmd += ["-u", to, branch]
    io.communicate(cmd, env=env, cwd=repo_path)


def branch(ctx, repo_path, branch, from_branch=None):
    env = git_env_for(repo_path)
    cmd = ["git", "checkout", "-b", branch]
    if from_branch is not None:
        cmd.append(from_branch)
    io.communicate(cmd, env=env)


def checkout(ctx, remote_repo, local_path, branch=None, remote="origin", from_branch="master"):
    """Checkout a new branch from a remote repository."""
    env = git_env_for(local_path)
    if not os.path.exists(local_path):
        io.communicate(command_clone(ctx, remote_repo, local_path))
    else:
        io.communicate(["git", "fetch", remote], env=env)

    if branch:
        io.communicate(["git", "checkout", f"{remote}/{from_branch}", "-b", branch], env=env)
    else:
        io.communicate(["git", "merge", "--ff-only", f"{remote}/{from_branch}"], env=env)


def command_clone(
    ctx: "PlanemoCliContext", src: str, dest: str, mirror: bool = False, branch: Optional[str] = None
) -> List[str]:
    """Produce a command-line string to clone a repository.

    Take in ``ctx`` to allow more configurability down the road.
    """
    cmd = ["git", "clone"]
    if mirror:
        cmd.append("--mirror")
    if branch is not None:
        cmd.extend(["--branch", branch])
    cmd.extend([src, dest])
    return cmd


def diff(ctx: "PlanemoCliContext", directory: str, range: str) -> List[str]:
    """Produce a list of diff-ed files for commit range."""
    cmd = f"cd '{directory}' && git diff --name-only '{range}' --"
    stdout, _ = io.communicate(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    return [line.strip() for line in unicodify(stdout).splitlines() if line]


def clone(*args, **kwds) -> None:
    """Clone a git repository.

    See :func:`command_clone` for description of arguments.
    """
    command = command_clone(*args, **kwds)
    io.communicate(command)


def rev(ctx: "PlanemoCliContext", directory: str) -> str:
    """Raw revision for git directory specified.

    Throws ``RuntimeError`` if not a git directory.
    """
    cmd = f"cd '{directory}' && git rev-parse HEAD"
    stdout, _ = io.communicate(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return unicodify(stdout).strip()


def is_rev_dirty(ctx: "PlanemoCliContext", directory: str) -> bool:
    """Check if specified git repository has uncommitted changes."""
    return io.shell(["git", "diff", "--quiet"], cwd=directory) != 0


def rev_if_git(ctx: "PlanemoCliContext", directory: str) -> Optional[str]:
    """Determine git revision (or ``None``)."""
    try:
        the_rev = rev(ctx, directory)
        is_dirty = is_rev_dirty(ctx, directory)
        if is_dirty:
            the_rev += "-dirty"
        return the_rev
    except RuntimeError:
        return None



"""Module contains code for the Tutorial class, dealing with the creation of a training tutorial."""

import collections
import json
import os
import re
import shutil

import oyaml as yaml
import requests

from planemo import templates
from planemo.bioblend import galaxy
from planemo.engine import (
    engine_context,
    is_galaxy_engine,
)
from planemo.io import (
    error,
    info,
)
from planemo.runnable import for_path
from .tool_input import (
    get_empty_input,
    get_empty_param,
    ToolInput,
)
from .utils import (
    load_yaml,
    save_to_yaml,
)

TUTO_HAND_ON_TEMPLATE = """---
layout: tutorial_hands_on

{{ metadata }}
---

{{ body }}
"""

TUTO_SLIDES_TEMPLATE = """---
layout: tutorial_slides
logo: "GTN"

{{ metadata }}
---

### How to fill the slide decks?

Please follow our
[tutorial to learn how to fill the slides]({{ '{{' }} site.baseurl {{ '}}' }}/topics/contributing/tutorials/create-new-tutorial-slides/slides.html)
"""


HANDS_ON_TOOL_BOX_TEMPLATE = """
## Sub-step with **{{tool_name}}**

> <hands-on-title> Task description </hands-on-title>
>
> 1. {{ '{%' }} tool [{{tool_name}}]({{tool_id}}) {{ '%}' }} with the following parameters:{{inputlist}}{{paramlist}}
>
>    ***TODO***: *Check parameter descriptions*
>
>    ***TODO***: *Consider adding a comment or tip box*
>
>    > <comment-title> short description </comment-title>
>    >
>    > A comment about the tool or something else. This box can also be in the main text
>    {: .comment}
>
{: .hands_on}

***TODO***: *Consider adding a question to test the learners understanding of the previous exercise*

> <question-title></question-title>
>
> 1. Question1?
> 2. Question2?
>
> > <solution-title></solution-title>
> >
> > 1. Answer for question1
> > 2. Answer for question2
> >
> {: .solution}
>
{: .question}

"""

TUTO_BIBLIOGRAPHY_TEMPLATE = """
# This is the bibliography file for your tutorial.
#
# To add bibliography (bibtex) entries here, follow these steps:
#  1) Find the DOI for the article you want to cite
#  2) Go to https://doi2bib.org and fill in the DOI
#  3) Copy the resulting bibtex entry into this file
#
# To cite the example below, in your tutorial.md file
# use {{ '{%' }} cite Batut2018 {{ '%}' }}
#
# If you want to cite an online resourse (website etc)
# you can use the 'online' format (see below)
#
# You can remove the examples below

@article{Batut2018,
  doi = {10.1016/j.cels.2018.05.012},
  url = {https://doi.org/10.1016/j.cels.2018.05.012},
  year = {2018},
  month = jun,
  publisher = {Elsevier {BV}},
  volume = {6},
  number = {6},
  pages = {752--758.e1},
  author = {B{\\'{e}}r{\\'{e}}nice Batut and Saskia Hiltemann and Andrea Bagnacani and Dannon Baker and Vivek Bhardwaj and
           Clemens Blank and Anthony Bretaudeau and Loraine Brillet-Gu{\\'{e}}guen and Martin {\\v{C}}ech and John Chilton
           and Dave Clements and Olivia Doppelt-Azeroual and Anika Erxleben and Mallory Ann Freeberg and Simon Gladman and
           Youri Hoogstrate and Hans-Rudolf Hotz and Torsten Houwaart and Pratik Jagtap and Delphine Larivi{\\`{e}}re and
           Gildas Le Corguill{\\'{e}} and Thomas Manke and Fabien Mareuil and Fidel Ram{\\'{i}}rez and Devon Ryan and
           Florian Christoph Sigloch and Nicola Soranzo and Joachim Wolff and Pavankumar Videm and Markus Wolfien and
           Aisanjiang Wubuli and Dilmurat Yusuf and James Taylor and Rolf Backofen and Anton Nekrutenko and Bj\\"{o}rn Gr\\"{u}ning},
  title = {Community-Driven Data Analysis Training for Biology},
  journal = {Cell Systems}
}

@online{gtn-website,
  author = {GTN community},
  title = {GTN Training Materials: Collection of tutorials developed and maintained by the worldwide Galaxy community},
  url = {https://training.galaxyproject.org},
  urldate = {2021-03-24}
}

"""

TUTO_HAND_ON_BODY_TEMPLATE = """
# Introduction

<!-- This is a comment. -->

General introduction about the topic and then an introduction of the
tutorial (the questions and the objectives). It is nice also to have a
scheme to sum up the pipeline used during the tutorial. The idea is to
give to trainees insight into the content of the tutorial and the (theoretical
and technical) key concepts they will learn.

You may want to cite some publications; this can be done by adding citations to the
bibliography file (`tutorial.bib` file next to your `tutorial.md` file). These citations
must be in bibtex format. If you have the DOI for the paper you wish to cite, you can
get the corresponding bibtex entry using [doi2bib.org](https://doi2bib.org).

With the example you will find in the `tutorial.bib` file, you can add a citation to
this article here in your tutorial like this:
{{ '{%' }} raw {{ '%}' }} `{{ '{%' }} cite Batut2018 {{ '%}' }}`{{ '{%' }} endraw {{ '%}' }}.
This will be rendered like this: {{ '{%' }} cite Batut2018 {{ '%}' }}, and links to a
[bibliography section](#bibliography) which will automatically be created at the end of the
tutorial.


**Please follow our
[tutorial to learn how to fill the Markdown]({{ '{{' }} site.baseurl {{ '}}' }}/topics/contributing/tutorials/\
create-new-tutorial-content/tutorial.html)**

> <agenda-title></agenda-title>
>
> In this tutorial, we will cover:
>
> 1. TOC
> {:toc}
>
{: .agenda}

# Title for your first section

Give some background about what the trainees will be doing in the section.
Remember that many people reading your materials will likely be novices,
so make sure to explain all the relevant concepts.

## Title for a subsection
Section and subsection titles will be displayed in the tutorial index on the left side of
the page, so try to make them informative and concise!

# Hands-on Sections
Below are a series of hand-on boxes, one for each tool in your workflow file.
Often you may wish to combine several boxes into one or make other adjustments such
as breaking the tutorial into sections, we encourage you to make such changes as you
see fit, this is just a starting point :)

Anywhere you find the word "***TODO***", there is something that needs to be changed
depending on the specifics of your tutorial.

have fun!

## Get data

> <hands-on-title> Data Upload </hands-on-title>
>
> 1. Create a new history for this tutorial
> 2. Import the files from [Zenodo]({{ '{{' }} page.zenodo_link {{ '}}' }}) or from
>    the shared data library (`GTN - Material` -> `{{ '{{' }} page.topic_name {{ '}}' }}`
>     -> `{{ '{{' }} page.title {{ '}}' }}`):
>
>    ```
>    {{ z_file_links }}
>    ```
>    ***TODO***: *Add the files by the ones on Zenodo here (if not added)*
>
>    ***TODO***: *Remove the useless files (if added)*
>
>    {{ '{%' }} snippet faqs/galaxy/datasets_import_via_link.md {{ '%}' }}
>
>    {{ '{%' }} snippet faqs/galaxy/datasets_import_from_data_library.md {{ '%}' }}
>
> 3. Rename the datasets
> 4. Check that the datatype
>
>    {{ '{%' }} snippet faqs/galaxy/datasets_change_datatype.md datatype="datatypes" {{ '%}' }}
>
> 5. Add to each database a tag corresponding to ...
>
>    {{ '{%' }} snippet faqs/galaxy/datasets_add_tag.md {{ '%}' }}
>
{: .hands_on}

# Title of the section usually corresponding to a big step in the analysis

It comes first a description of the step: some background and some theory.
Some image can be added there to support the theory explanation:

![Alternative text](../../images/image_name "Legend of the image")

The idea is to keep the theory description before quite simple to focus more on the practical part.

***TODO***: *Consider adding a detail box to expand the theory*

> <details-title> More details about the theory </details-title>
>
> But to describe more details, it is possible to use the detail boxes which are expandable
>
{: .details}

A big step can have several subsections or sub steps:

{{ body }}

## Re-arrange

To create the template, each step of the workflow had its own subsection.

***TODO***: *Re-arrange the generated subsections into sections or other subsections.
Consider merging some hands-on boxes to have a meaningful flow of the analyses*

# Conclusion

Sum up the tutorial and the key takeaways here. We encourage adding an overview image of the
pipeline used.
"""


class Tutorial:
    """Class to describe a training tutorial."""

    def __init__(self, training, topic, name="new_tuto", title="The new tutorial", zenodo_link=""):
        """Init a tutorial instance."""
        self.training = training
        self.topic = topic
        self.name = name
        self.title = title
        self.zenodo_link = zenodo_link
        self.zenodo_file_links = []
        self.questions = []
        self.objectives = []
        self.time = ""
        self.key_points = []
        self.contributors = []
        self.body = ""
        self.init_wf_fp = None
        self.init_wf_id = None
        self.hands_on = True
        self.slides = False
        self.datatype_fp = ""
        self.set_dir_name()
        self.init_data_lib()
        self.body = templates.render(
            HANDS_ON_TOOL_BOX_TEMPLATE,
            **{"tool_name": "My Tool", "inputlist": get_empty_input(), "paramlist": get_empty_param()},
        )

    def init_from_kwds(self, kwds):
        """Init a tutorial instance from a kwds dictionary."""
        self.name = kwds["tutorial_name"]
        self.title = kwds["tutorial_title"]
        self.zenodo_link = kwds["zenodo_link"] if kwds["zenodo_link"] else ""
        self.questions = [
            "Which biological questions are addressed by the tutorial?",
            "Which bioinformatics techniques are important to know for this type of data?",
        ]
        self.objectives = [
            "The learning objectives are the goals of the tutorial",
            "They will be informed by your audience and will communicate to them and to yourself what you should focus on during the course",
            "They are single sentences describing what a learner should be able to do once they have completed the tutorial",
            "You can use Bloom's Taxonomy to write effective learning objectives",
        ]
        self.time = "3H"
        self.key_points = ["The take-home messages", "They will appear at the end of the tutorial"]
        self.contributors = ["contributor1", "contributor2"]
        self.init_wf_fp = kwds["workflow"]
        self.init_wf_id = kwds["workflow_id"]
        self.hands_on = kwds["hands_on"]
        self.slides = kwds["slides"]
        self.datatype_fp = kwds["datatypes"]
        self.set_dir_name()
        self.init_data_lib()

    def init_from_existing_tutorial(self, tuto_name):
        """Init a tutorial instance from an existing tutorial (data library and tutorial.md)."""
        self.name = tuto_name
        self.set_dir_name()

        if not self.exists():
            raise Exception("The tutorial %s does not exists. It should be created" % self.name)

        # get the metadata information of the tutorial (from the top of the tutorial.md)
        with open(self.tuto_fp) as tuto_f:
            tuto_content = tuto_f.read()
        regex = r"^---\n(?P<metadata>[\s\S]*)\n---(?P<body>[\s\S]*)"
        tuto_split_regex = re.search(regex, tuto_content)
        if not tuto_split_regex:
            raise Exception("No metadata found at the top of the tutorial")
        metadata = yaml.safe_load(tuto_split_regex.group("metadata"))
        self.title = metadata["title"]
        self.zenodo_link = metadata["zenodo_link"]
        self.questions = metadata["questions"]
        self.objectives = metadata["objectives"]
        self.time_estimation = metadata["time_estimation"]
        self.key_points = metadata["key_points"]
        self.contributors = metadata["contributors"]

        # the tutorial content
        self.body = tuto_split_regex.group("body")

        # get the data library
        self.init_data_lib()

    def init_data_lib(self):
        """Init the data library dictionary."""
        if os.path.exists(self.data_lib_fp):
            self.data_lib = load_yaml(self.data_lib_fp)
        else:
            self.data_lib = collections.OrderedDict()
        # set default information
        self.data_lib.setdefault("destination", collections.OrderedDict())
        self.data_lib["destination"]["type"] = "library"
        self.data_lib["destination"]["name"] = "GTN - Material"
        self.data_lib["destination"]["description"] = "Galaxy Training Network Material"
        self.data_lib["destination"][
            "synopsis"
        ] = "Galaxy Training Network Material. See https://training.galaxyproject.org"
        self.data_lib.setdefault("items", [])
        self.data_lib.pop("libraries", None)
        # get topic or create new one
        topic = collections.OrderedDict()
        for item in self.data_lib["items"]:
            if item["name"] == self.topic.title:
                topic = item
        if not topic:
            self.data_lib["items"].append(topic)
            topic["name"] = self.topic.title
            topic["description"] = self.topic.summary
            topic["items"] = []
        # get tutorial or create new one
        self.tuto_data_lib = collections.OrderedDict()
        for item in topic["items"]:
            if item["name"] == self.title:
                self.tuto_data_lib = item
        if not self.tuto_data_lib:
            topic["items"].append(self.tuto_data_lib)
            self.tuto_data_lib["name"] = self.title
            self.tuto_data_lib["items"] = []

    # GETTERS
    def get_tuto_metata(self):
        """Return the string corresponding to the tutorial metadata."""
        metadata = collections.OrderedDict()
        metadata["title"] = self.title
        metadata["zenodo_link"] = self.zenodo_link
        metadata["questions"] = self.questions
        metadata["objectives"] = self.objectives
        metadata["time_estimation"] = self.time
        metadata["key_points"] = self.key_points
        metadata["contributors"] = self.contributors
        return yaml.safe_dump(metadata, indent=2, default_flow_style=False, default_style="", explicit_start=False)

    # SETTERS
    def set_dir_name(self):
        """Set the path to dir and files of a tutorial."""
        self.dir = os.path.join(self.topic.dir, "tutorials", self.name)
        self.tuto_fp = os.path.join(self.dir, "tutorial.md")
        self.bib_fp = os.path.join(self.dir, "tutorial.bib")
        self.slide_fp = os.path.join(self.dir, "slides.html")
        self.data_lib_fp = os.path.join(self.dir, "data-library.yaml")
        self.wf_dir = os.path.join(self.dir, "workflows")
        self.wf_fp = os.path.join(self.wf_dir, "main_workflow.ga")
        self.faq_dir = os.path.join(self.dir, "faqs")
        self.tour_dir = os.path.join(self.dir, "tours")
        # remove empty workflow file if there
        empty_wf_filepath = os.path.join(self.wf_dir, "empty_workflow.ga")
        if os.path.exists(empty_wf_filepath):
            os.remove(empty_wf_filepath)

    # TEST METHODS
    def exists(self):
        """Test if the tutorial exists."""
        return os.path.isdir(self.dir)

    def has_workflow(self):
        """Test if a workflow is provided for the tutorial."""
        return self.init_wf_fp or self.init_wf_id

    # EXPORT METHODS
    def export_workflow_file(self):
        """Copy or extract workflow file and add it to the tutorial directory."""
        if not os.path.exists(self.wf_dir):
            os.makedirs(self.wf_dir)
        if not os.path.exists(os.path.join(self.wf_dir, "index.md")):
            with open(os.path.join(self.wf_dir, "index.md"), "w") as handle:
                handle.write("---\nlayout: workflow-list\n---\n")
        if self.init_wf_fp:
            shutil.copy(self.init_wf_fp, self.wf_fp)
        elif self.init_wf_id:
            gi = galaxy.GalaxyInstance(self.training.galaxy_url, key=self.training.galaxy_api_key)
            gi.workflows.export_workflow_to_local_path(self.init_wf_id, self.wf_fp, use_default_filename=False)

    # OTHER METHODS
    def get_files_from_zenodo(self):
        """Extract a list of URLs and dictionary describing the files from the JSON output of the Zenodo API."""
        z_record, req_res = get_zenodo_record(self.zenodo_link)

        self.zenodo_file_links = []
        if "files" not in req_res:
            raise ValueError("No files in the Zenodo record")

        files = []
        for f in req_res["files"]:
            file_dict = {"url": "", "src": "url", "ext": "", "info": self.zenodo_link}
            if "type" in f:
                file_dict["ext"] = get_galaxy_datatype(f["type"], self.datatype_fp)
            if "links" not in f and "self" not in f["links"]:
                raise ValueError("No link for file %s" % f)
            file_dict["url"] = f["links"]["self"]
            self.zenodo_file_links.append(f["links"]["self"])
            files.append(file_dict)

        return (files, z_record)

    def prepare_data_library_from_zenodo(self):
        """Get the list of URLs of the files on Zenodo, fill the data library, save it into the file."""
        self.zenodo_file_links = []
        if self.zenodo_link != "":
            files, z_record = self.get_files_from_zenodo()
            if z_record:
                # get current data library and/or previous data library for the tutorial
                # remove the latest tag of any existing library
                # remove the any other existing library
                current_data_lib = collections.OrderedDict()
                previous_data_lib = collections.OrderedDict()
                for item in self.tuto_data_lib["items"]:
                    if item["name"] == "DOI: 10.5281/zenodo.%s" % z_record:
                        current_data_lib = item
                    elif item["description"] == "latest":
                        previous_data_lib = item
                        previous_data_lib["description"] = ""
                if not current_data_lib:
                    current_data_lib["name"] = "DOI: 10.5281/zenodo.%s" % z_record
                    current_data_lib["description"] = "latest"
                    current_data_lib["items"] = []
                current_data_lib["items"] = files

                self.tuto_data_lib["items"] = [current_data_lib]
                if previous_data_lib:
                    self.tuto_data_lib["items"].append(previous_data_lib)
        save_to_yaml(self.data_lib, self.data_lib_fp)

    def write_hands_on_tutorial(self, add_z_file_links=True):
        """Write the content of the hands-on tutorial in the corresponding file."""
        if add_z_file_links:
            self.body = templates.render(
                TUTO_HAND_ON_BODY_TEMPLATE,
                **{"z_file_links": "\n>    ".join(self.zenodo_file_links), "body": self.body},
            )
        # write in the tutorial file with the metadata on the top
        metadata = self.get_tuto_metata()
        with open(self.tuto_fp, "w") as md:
            md.write(templates.render(TUTO_HAND_ON_TEMPLATE, **{"metadata": metadata, "body": self.body}))

        # create the bibliography file
        self.write_bibliography()

    def write_bibliography(self):
        """Write the content of the bibliography file for the tutorial."""
        with open(self.bib_fp, "w") as bib:
            bib.write(templates.render(TUTO_BIBLIOGRAPHY_TEMPLATE, **{"body": self.body}))

    def create_hands_on_tutorial(self, ctx):
        """Create tutorial structure from the workflow file (if it is provided)."""
        # load workflow and get hands-on body from the workflow
        if self.init_wf_id:
            if not self.training.galaxy_url:
                raise ValueError("No Galaxy URL given")
            if not self.training.galaxy_api_key:
                raise ValueError("No API key to access the given Galaxy instance")
            self.body = get_hands_on_boxes_from_running_galaxy(
                self.init_wf_id, self.training.galaxy_url, self.training.galaxy_api_key
            )
        elif self.init_wf_fp:
            self.body = get_hands_on_boxes_from_local_galaxy(self.training.kwds, self.init_wf_fp, ctx)
        # write tutorial body
        self.write_hands_on_tutorial()

    def create_tutorial(self, ctx):
        """Create the skeleton of a new tutorial."""
        # create tuto folder and empty files
        os.makedirs(self.dir)
        os.makedirs(self.tour_dir)
        os.makedirs(self.wf_dir)

        # extract the data library from Zenodo and the links for the tutorial
        if self.zenodo_link != "":
            info("Create the data library from Zenodo")
            self.prepare_data_library_from_zenodo()

        # create tutorial skeleton from workflow and copy workflow file
        if self.hands_on:
            info("Create tutorial skeleton from workflow (if it is provided)")
            self.create_hands_on_tutorial(ctx)
            self.export_workflow_file()

        # create slide skeleton
        if self.slides:
            with open(self.slide_fp, "w") as slide_f:
                slide_f.write(templates.render(TUTO_SLIDES_TEMPLATE, **{"metadata": self.get_tuto_metata()}))

        # create the FAQ page
        os.makedirs(self.faq_dir)
        if not os.path.exists(os.path.join(self.faq_dir, "index.md")):
            with open(os.path.join(self.faq_dir, "index.md"), "w") as handle:
                handle.write("---\nlayout: faq-page\n---\n")


def get_galaxy_datatype(z_ext, datatype_fp):
    """Get the Galaxy datatype corresponding to a Zenodo file type."""
    g_datatype = ""
    datatypes = load_yaml(datatype_fp)
    if z_ext in datatypes:
        g_datatype = datatypes[z_ext]
    if g_datatype == "":
        g_datatype = "# Please add a Galaxy datatype or update the shared/datatypes.yaml file"
    info(f"Get Galaxy datatypes: {z_ext} --> {g_datatype}")
    return g_datatype


def get_zenodo_record(zenodo_link):
    """Get the content of a Zenodo record."""
    # get the record in the Zenodo link
    if "doi" in zenodo_link:
        z_record = zenodo_link.split(".")[-1]
    else:
        z_record = zenodo_link.split("/")[-1]
    # get JSON corresponding to the record from Zenodo API
    req = "https://zenodo.org/api/records/%s" % (z_record)
    try:
        r = requests.get(req)
        r.raise_for_status()
        req_res = r.json()
    except Exception as e:
        error(f"The Zenodo link ({zenodo_link}) seems invalid: {e}")
        req_res = {"files": []}
        z_record = None
    return (z_record, req_res)


def get_wf_inputs(step_inp):
    """Get the inputs from a workflow step and format them into a hierarchical dictionary."""
    inputs = {}
    for inp_n, inp in step_inp.items():
        if "|" in inp_n:
            repeat_regex = r"(?P<prefix>[^\|]*)_(?P<nb>\d+)\|(?P<suffix>.+).+"
            repeat_search = re.search(repeat_regex, inp_n)
            hier_regex = r"(?P<prefix>[^\|]*)\|(?P<suffix>.+)"
            hier_regex = re.search(hier_regex, inp_n)
            if repeat_search and repeat_search.start(0) <= hier_regex.start(0):
                inputs.setdefault(repeat_search.group("prefix"), {})
                inputs[repeat_search.group("prefix")].setdefault(
                    repeat_search.group("nb"), get_wf_inputs({hier_regex.group("suffix"): inp})
                )
            else:
                inputs.setdefault(hier_regex.group("prefix"), {})
                inputs[hier_regex.group("prefix")].update(get_wf_inputs({hier_regex.group("suffix"): inp}))
        else:
            inputs.setdefault(inp_n, inp)
    return inputs


def get_wf_param_values(init_params, inp_connections):
    """Get the param values from a workflow step and format them into a hierarchical dictionary."""
    if not isinstance(init_params, str) or '": ' not in init_params:
        form_params = init_params
    else:
        form_params = json.loads(init_params)
    if isinstance(form_params, dict):
        if "__class__" in form_params and (
            form_params["__class__"] == "RuntimeValue" or form_params["__class__"] == "ConnectedValue"
        ):
            form_params = inp_connections
        else:
            for p in form_params:
                inp = inp_connections[p] if p in inp_connections else {}
                form_params[p] = get_wf_param_values(form_params[p], inp)
    elif isinstance(form_params, list):
        json_params = form_params
        form_params = []
        for i, p in enumerate(json_params):
            inp = inp_connections[str(i)] if str(i) in inp_connections else {}
            form_params.append(get_wf_param_values(p, inp))
    elif isinstance(form_params, str) and '"' in form_params:
        form_params = form_params.replace('"', "")
    return form_params


def format_wf_steps(wf, gi):
    """Get a string with the hands-on boxes describing the different steps of the worklow."""
    body = ""
    steps = wf["steps"]

    for s in range(len(steps)):
        wf_step = steps[str(s)]

        # get params in workflow
        wf_param_values = {}
        if wf_step["tool_state"] and wf_step["input_connections"]:
            wf_param_values = get_wf_param_values(wf_step["tool_state"], get_wf_inputs(wf_step["input_connections"]))
        if not wf_param_values:
            continue
        # get tool description
        try:
            tool_desc = gi.tools.show_tool(wf_step["tool_id"], io_details=True)
        except Exception:
            tool_desc = {"inputs": []}
        # get formatted param description
        paramlist = ""
        for inp in tool_desc["inputs"]:
            if inp["name"].startswith("__"):
                continue
            tool_inp = ToolInput(inp, wf_param_values, steps, 1, should_be_there=True)
            paramlist += tool_inp.get_formatted_desc()
        # format the hands-on box
        body += templates.render(
            HANDS_ON_TOOL_BOX_TEMPLATE,
            **{"tool_name": wf_step["name"], "tool_id": wf_step["tool_id"], "paramlist": paramlist},
        )
    return body


def get_hands_on_boxes_from_local_galaxy(kwds, wf_filepath, ctx):
    """Server local Galaxy and get the workflow dictionary."""
    assert is_galaxy_engine(**kwds)
    runnable = for_path(wf_filepath)
    tuto_body = ""
    with engine_context(ctx, **kwds) as galaxy_engine:
        with galaxy_engine.ensure_runnables_served([runnable]) as config:
            info("Status of installed repositories: %s" % config.gi.toolshed.get_repositories())
            workflow_id = config.workflow_id(wf_filepath)
            wf = config.gi.workflows.export_workflow_dict(workflow_id)
            tuto_body = format_wf_steps(wf, config.gi)
    return tuto_body


def get_hands_on_boxes_from_running_galaxy(wf_id, galaxy_url, galaxy_api_key):
    """Get the workflow dictionary from a running Galaxy instance with the workflow installed on it."""
    gi = galaxy.GalaxyInstance(galaxy_url, key=galaxy_api_key)
    wf = gi.workflows.export_workflow_dict(wf_id)
    tuto_body = format_wf_steps(wf, gi)
    return tuto_body



"""Module describing the planemo ``training_generate_from_wf`` command."""

import click

from planemo import options
from planemo.cli import command_function
from planemo.training import Training


@click.command("training_generate_from_wf")
@options.optional_tools_arg(multiple=True, allow_uris=True)
@options.training_generate_tuto_from_wf_options()
@options.galaxy_serve_options()
@command_function
def cli(ctx, uris, **kwds):
    """Create tutorial skeleton from workflow."""
    kwds["no_dependency_resolution"] = True
    # Ugh rather than override this - it just needs to not be in the serve options
    # for this command.
    kwds["skip_client_build"] = True
    training = Training(kwds)
    training.generate_tuto_from_wf(ctx)



"""Module describing the planemo ``shed_test`` command."""
import sys

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function
from planemo.engine import engine_context
from planemo.galaxy.test.actions import handle_reports_and_summary
from planemo.network_util import get_free_port
from planemo.runnable_resolve import install_args_list_to_runnables


@click.command("shed_test")
@options.shed_read_options()
@options.galaxy_target_options()
@options.test_options()
@click.option(
    "--skip_dependencies", is_flag=True, help="Do not install shed dependencies as part of repository installation."
)
@command_function
def cli(ctx, paths, **kwds):
    """Run tests of published shed artifacts.

    This command will start a Galaxy instance configured to target the
    specified shed, find published artifacts (tools and dependencies)
    corresponding to command-line arguments and ``.shed.yml`` file(s),
    install these artifacts, and run the tool tests for these commands.

    This command requires the target to be version 15.07 or newer.
    """
    install_args_list = kwds["install_args_list"] = shed.install_arg_lists(ctx, paths, **kwds)
    runnables = install_args_list_to_runnables(ctx, install_args_list, kwds)
    kwds["port"] = get_free_port()
    with engine_context(ctx, **kwds) as engine:
        test_data = engine.test(runnables, test_timeout=kwds.get("test_timeout"))
        ctx.vlog(f"engine.test returning [{test_data}]")
        return_code = handle_reports_and_summary(ctx, test_data.structured_data, kwds=kwds)
    sys.exit(return_code)
